"""Utility functions for monitors"""

import curses
import logging
from math import floor
from typing import Tuple, Type

# Getting the logger reference
logger = logging.getLogger()


def logging_console(enabled: bool = True) -> None:
    """
    Disable logger standard out in terminal, only keep log file output

    Args:
        None

    Returns:
        None
    """
    # FIXME: Instead of blindly indexing a handler, find/match the right one

    # [ handler.setLevel(logging.FATAL) for handler in logger.handlers if isinstance(handler, logging.StreamHandler) ]
    # for handler in logger.handlers:
    #     if isinstance(handler, logging.StreamHandler):
    #         handler.setLevel(logging.FATAL)

    logger.debug('Logging handler status:')
    for i, handler in enumerate(logger.handlers):
        logger.debug(f'    {i + 1}. {type(handler)} - Logging Level: {logging.getLevelName(handler.level)}')

    if enabled:
        logger.debug('*******************************************')
        logger.debug('***  LOGGING TO CONSOLE OUTPUT ENABLED  ***')
        logger.debug('*******************************************')
        stream_handler = logger.handlers[1]
        stream_handler.setLevel(logging.DEBUG)
    else:
        logger.debug('**********************************************************************')
        logger.debug('***  LOGGING TO CONSOLE OUTPUT DISABLED. ONLY LOGGING TO LOG FILE  ***')
        logger.debug('**********************************************************************')
        stream_handler = logger.handlers[1]
        stream_handler.setLevel(logging.FATAL)

    logger.debug('Logging handler status:')
    for i, handler in enumerate(logger.handlers):
        logger.debug(f'    {i + 1}. {type(handler)} - Logging Level: {logging.getLevelName(handler.level)}')


def load_curses_colors_decor() -> Tuple[dict, dict]:
    """
    Load curses colors and decorations and load them in a usable
    dictionary for reference.

    Usage: self.colors("green")
        self.decor("blink")
        scr.addstr(y, x, "hello", self.color['blue'] | self.decor['bold'])

    Args:
        None

    Returns:
        color (dict): Curses color references
        decor (dict): Curses decoration/style references
    """
    # Start colors in curses
    curses.start_color()

    # Default background color
    bkgrd_color = 0

    # Defining colors [foreground/font, background]
    color_definition = {
        'normal': [curses.COLOR_WHITE, bkgrd_color],  # FIXME: Rename to "normal" to match decor
        'red': [curses.COLOR_RED, bkgrd_color],
        'green': [curses.COLOR_GREEN, bkgrd_color],
        'blue': [curses.COLOR_BLUE, bkgrd_color],
        'yellow': [curses.COLOR_YELLOW, bkgrd_color],
        'orange': [209, bkgrd_color],
        'cyan': [curses.COLOR_CYAN, bkgrd_color],
        'magenta': [curses.COLOR_MAGENTA, bkgrd_color],
        'grey-dark': [240, bkgrd_color],
        'grey-light': [248, bkgrd_color],
        'black-white': [curses.COLOR_BLACK, curses.COLOR_WHITE],
        'white-red': [curses.COLOR_WHITE, curses.COLOR_RED]
    }

    # If terminal does not support colors, reset everything to white
    if not curses.has_colors() or not curses.can_change_color():
        for color in color_definition:
            color_definition[color] = [curses.COLOR_WHITE, bkgrd_color]

    # Initiating curses color and saving for quick reference
    color = {}
    for index, (key, value) in enumerate(color_definition.items()):
        try:
            curses.init_pair(index + 1, value[0], value[1])
        except:
            curses.init_pair(index + 1, 0, 0)
        color[key] = curses.color_pair(index + 1)

    # Defining font decorations
    decoration_definition = {
        'normal': curses.A_NORMAL,  # Normal display (no highlight)
        'standout': curses.A_STANDOUT,  # Best highlighting mode of the terminal
        'underline': curses.A_UNDERLINE,  # Underlining
        'reverse': curses.A_REVERSE,  # Reverse video
        'blink': curses.A_BLINK,  # Blinking
        'dim': curses.A_DIM,  # Half bright
        'bold': curses.A_BOLD,  # Extra bright or bold
        'protect': curses.A_PROTECT,  # Protected mode
        'invisible': curses.A_INVIS,  # Invisible or blank mode
        'alt-char': curses.A_ALTCHARSET,  # Alternate character set
        'char':
            curses.A_CHARTEXT  # Bit-mask to extract a character
    }

    # Initiating curses color and saving for quick reference
    decor = {}
    for key, value in decoration_definition.items():
        decor[key] = value

    return color, decor


def load_keys() -> dict:
    """
    Load all keyboard keys available to user in program

    Usage: ui_keys['DOWN']

    Args:
        None
    Returns:
        ui_keys (dict): Dictionary of references to curses keys
    """
    ui_keys = {
        "ABORT": (ord('a'), ord('A')),
        "BUILD": (ord('b'), ord('B')),
        "DOWN": (curses.KEY_DOWN, ord('j')),
        "ENTER": (curses.KEY_ENTER, ord('\n'), ord('\r')),
        "HELP": (ord('h'), ord('H')),
        "LEFT": (curses.KEY_LEFT, ord('h')),
        "LOGS": (ord('l'), ord('L')),
        "OPEN": (ord('o'), ord('O')),
        "PAUSE": (ord('p'), ord('P')),
        "QUIT": (27, ord('q'), ord('Q')),
        "RESUME": (ord('r'), ord('R')),
        "RIGHT": (curses.KEY_RIGHT, ord('l')),
        "SOUND": (ord('s'), ord('S')),
        "SPACE": (32, ord(' ')),
        "UP": (curses.KEY_UP, ord('k'))
    }
    return ui_keys


def get_center_x(scr, line: str) -> int:
    """
    Find the horizontal center position of the given text line

    Args:
        display_width (int) : The character width of the screen/space/display
        line (int)          : Line of text
    Returns:
        (int): Horizontal character number
    """
    # Getting the screen height and width
    _, term_width = scr.getmaxyx()
    return term_width // 2 - len(line) // 2


def get_center_y(scr) -> int:
    """
    Find the vertical center position of given screen/space/display

    Args:
        display_width (int) : The character height of the screen/space/display
    Returns:
        (int): Vertical character number
    """
    # Getting the screen height and width
    term_height, _ = scr.getmaxyx()
    return term_height // 2


def truncate_text(text: str, length_limit: int) -> str:
    """
    Truncating/shortening of text given a length limit.
    Will add "..." to truncated text.

    Args:
        text (str)         : The character height of the screen/space/display
        length_limit (int) : Length limit of the text
    Returns:
        truncated_text (str): The truncated line of text
    """
    truncated_text = text
    if len(text) >= length_limit - 3:
        truncated_text = text[0:length_limit - 3] + '...'

    return truncated_text


def get_message_box_size(term_height: int, term_width: int, message_lines: list) -> Tuple[int, int, int, int]:
    """
    Given a message box list with each item being a message box line/row,
    this method find the right size and position of the message box for
    the given terminal size

    Args:
        term_height (int)    : Number of rows/lines in terminal
        term_width (int)     : Number of columns in terminal
        message_lines (list) : Lines of text in each list item
    Returns:
        box_height (int) : Height of message box (rows/lines)
        box_width (int)  : Width of message box (columns)
        box_y (int)      : Vertical position of box in terminal
        box_x (int)      : Horizontal position of box in terminal
    """
    box_height = len(message_lines) + 4
    box_width = int(term_width / 1.5)  # Alternative: len(max(message_lines, key=len)) + 12
    box_y = term_height // 2 - box_height // 2
    box_x = term_width // 2 - box_width // 2

    return box_height, box_width, box_y, box_x


def get_progress_bar(exam_progress: float, bar_char_width=60, bar_char_full='|', bar_char_empty='-') -> str:
    """
    Make a progress bar with specified parameters

    Args:
        exam_progress (float) : Exam progress from 0 to 1 (ie. 0.45 is 45%)
        bar_char_width (int)  : Total width of progress bar, columns of text
        bar_char_full (str)   : Symbol for filled
        bar_char_empty (str)  : Symbol for empty
    Returns:
        (str) : Progress bar as text
    """
    # TODO: Different colors for different parts of the progress bar somehow

    progress_str = []
    for i in range(bar_char_width):

        if i <= exam_progress * bar_char_width:
            progress_str.append(bar_char_full)
        else:
            progress_str.append(bar_char_empty)

    progress_str = "".join(progress_str)
    return progress_str


def draw_screen_border(scr, color: list) -> None:
    """
    Draw a border around entire terminal screen with specified color

    Parameters:
        scr (obj)   : Handle for curses terminal screen handle
        color (list): Foreground and background color (ie. [250, 0])
    Returns:
        None
    """
    scr.attron(color)
    scr.border(0)
    scr.attroff(color)


def draw_horizontal_header(scr,
                           row: int,
                           col: int,
                           width: int,
                           symbol: str = '-',
                           text: str = '',
                           color: list = []) -> None:
    """
    Draw a horizontal header on the screen
    Example:  `---------  HELLO  -------------`

    Args:
        scr    : Handle for curses terminal screen handle
        col    : What character start the seperator along the horizontal
        row    : The line/row number from top of the screen
        width  : Total width of the header
        symbol : (Optional) The symbol used on each side of the text (default: `-`)
        text   : (Optional) Text in center of seperator. Must be smaller than `width` (default: None)
        color  : (Optional) Foreground and background color (ie. `[250, 0]`)

    Returns:
        None
    """
    # Getting the screen height and width
    term_height, term_width = scr.getmaxyx()

    # Check text size and header width size
    if len(text) > width - 2 or width > term_width:
        return

    # Check if drawing within boundaries and if text is smaller than the width and some padding
    if row < term_height - 2 and row > 1:
        if text:
            # Get one half of the seperator
            seperator_half = symbol * (floor((width - len(text)) / 2) - 2)
            line = seperator_half + '  ' + text + '  ' + seperator_half
        else:
            line = symbol * width
        scr.addstr(row, col, line, color)


def draw_horizontal_seperator(scr, row: int, color: list = [], symbol: str = '-', text: str = '') -> None:
    """
    Draw a horizontal line accross the terminal screen at specified height
    and with specified color

    Args:
        scr (obj)   : Handle for curses terminal screen handle
        row (int)   : The line/row number from top of the screen
        color       : (Optional) Foreground and background color (ie. [250, 0])
        symbol      : (Optional) The symbol used as the seperator
        text        : (Optional) Text in center of seperator
    Returns:
        None
    """
    # Getting the screen height and width
    term_height, term_width = scr.getmaxyx()

    # Check if drawing within boundaries
    if row < term_height - 2 and row > 1:
        if text:
            # Get one half of the seperator
            seperator_half = symbol * (floor((term_width - len(text)) / 2) - 2)
            line = seperator_half + '  ' + text + '  ' + seperator_half
        else:
            line = symbol * term_width
        scr.addstr(row, 1, line, color)


def draw_vertical_seperator(scr, x: int, color: list) -> None:
    """
    Draw a vertical line accross the terminal screen at specified character
    column and with specified color

    Args:
        scr (obj)   : Handle for curses terminal screen handle
        x (int)     : The column number from left of the screen
        color (list): Foreground and background color (ie. [250, 0])
    Returns:
        None
    """
    # TODO: Currently unused but may come in handy


def draw_message_box(scr, message_lines: list, justify: str = 'center') -> None:
    """
    Draw a message box in the middle of the terminal screen

    Args:
        scr (obj)            : Handle for curses terminal screen handle
        message_lines (list) : Lines of text in each list item
        justify              : Text justification. `center`, `left`
    Returns:
        None
    """
    term_height, term_width = scr.getmaxyx()

    # Getting the message box size and position and creating the message box
    height, width, y, x = get_message_box_size(term_height, term_width, message_lines)
    message_box = curses.newwin(height, width, y, x)
    message_box.box()
    message_box.border()

    # Find the longest line of text
    max_line_len = len(max(message_lines, key=len))

    for l, line in enumerate(message_lines):
        # Getting the horizontal starting point
        if justify == 'center':
            x = width // 2 - len(line) // 2
        elif justify == 'left':
            x = width // 2 - max_line_len // 2

        # Add text each line to box (Text is relative to box -> y, x)
        y = l + 2
        message_box.addstr(y, x, line)

    # Refresh the message box
    message_box.refresh()


def draw_text(scr,
              text: str = None,
              y: int = None,
              x: int = None,
              color: list = None,
              decor: int = None,
              center_y: bool = False,
              center_x: bool = False) -> None:
    """
    Draw a text on screen.
    This is a wrapper for `addstr`

    Args:
        scr      : Handle for curses terminal screen handle
        text     : Lines of text in each list item
        y        : Starting character row in terminal
        x        : Starting character column in terminal
        color    : Color of text (0 - 255)
        decor    : Decoration of text
        center_y : Vertically center the text on screen
        center_x : Horizontally center the text on screen
    Returns:
        None
    """
    # Set to default color and decor if not passed
    if not color or not decor:
        color_preset, decor_preset = load_curses_colors_decor()
    color = color_preset['normal'] if not color else color
    decor = decor_preset['normal'] if not decor else decor

    # Check for NoneType
    if text == None:
        text = 'N/A'

    # Get center of screen if specified
    if center_x and not x:
        x = get_center_x(scr, text)
    if center_y and not y:
        y = get_center_y(scr)

    # Draw
    scr.addstr(y, x, text, color | decor)

    # Update but don't write yet
    scr.noutrefresh()


def paint_background(scr, color: int = 0) -> None:
    """
    Paint the background of the terminal window with one color

    Args:
        scr   : Handle for curses terminal screen handle
        color : Color index (0 - 255) (default: 0 - black)
    Returns:
        None
    """
    # Getting the screen height and width
    term_height, term_width = scr.getmaxyx()

    # Paint it
    for start_y in range(1, term_height - 1):
        line = term_width * ' '
        scr.addstr(start_y, 1, line, color)

    # Update but don't write yet
    scr.noutrefresh()



"""Jenkins Shared Library Management"""

import logging
import os

from yojenkins.utility import utility

# Getting the logger reference
logger = logging.getLogger()


class SharedLibrary():
    """Class managing a Jenkins Shared Library"""

    def __init__(self) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.groovy_script_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'groovy_scripts')

    def setup(self, rest: object, lib_name: str, repo_owner: str, repo_name: str, repo_url: str, repo_branch: str,
              implicit: bool, credential_id: str) -> bool:
        """TODO Docstring

        Details: TODO

        Args:
            TODO

        Returns:
            TODO
        """
        if not repo_url and not repo_owner and not repo_name:
            logger.error('Either git repository URL OR owner and name are required')
            return False

        if not repo_url:
            if repo_owner and not repo_name:
                logger.debug("Git repository name is missing")
                return False
            elif repo_name and not repo_owner:
                logger.debug("Git repository owner is missing")
                return False
            logger.debug("Using git owner/organization")
        else:
            logger.debug("Using git repository URL")

        logger.debug("Setting up shared library ...")
        logger.debug(f"   - Name:              {lib_name}")
        logger.debug(f"   - Repository URL:    {repo_url}")
        logger.debug(f"   - Repository Owner:  {repo_owner}")
        logger.debug(f"   - Repository Name:   {repo_name}")
        logger.debug(f"   - Repository Branch: {repo_branch}")
        logger.debug(f"   - Implicit Loading:  {implicit}")
        logger.debug(f"   - Credential ID:     {credential_id}")

        kwargs = {
            'lib_name': lib_name,
            'repo_owner': repo_owner,
            'repo_name': repo_name,
            'repo_url': repo_url,
            'repo_branch': repo_branch,
            'implicit': 'true' if implicit else 'false',
            'credential_id': credential_id
        }
        script_filepath = os.path.join(self.groovy_script_directory, 'shared_lib_setup.groovy')
        _, success, error = utility.run_groovy_script(script_filepath=script_filepath,
                                                      json_return=False,
                                                      rest=rest,
                                                      **kwargs)
        if not success:
            logger.debug(f"Failed to setup Jenkins shared library. {error}")
            return False
        return True



"""Monitor parent class"""

import curses
import logging
import platform
import sys
import threading
from time import sleep, time

if platform.system() != "Windows":
    try:
        import simpleaudio
    except:
        pass
else:
    import winsound

from yojenkins.yo_jenkins.status import Color, Sound, Status

from . import monitor_utility as mu

# Getting the logger reference
logger = logging.getLogger()


class Monitor:
    """Parent class for all monitor objects"""

    def __init__(self) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.rest = None
        self.auth = None
        self.job = None
        self.build = None

        self.color = {}
        self.decor = {}

        self.halfdelay_screen_refresh = 5  # number of 10th of a second

        self.height_limit = 35
        self.width_limit = 66
        self.terminal_size_good = True

        self.curses_alive = True

        self.quit = 0
        self.exit = False
        self.help = False

        self.server_status_data = {}
        self.server_status_thread_interval = 0.0

        self.playing_sound = False

        self.all_threads_enabled = True
        self.paused = False

        self.server_interaction = False

    def __del__(self):
        """ TODO """
        # Just in case turn off all threads
        self.all_threads_off()

    ###########################################################################
    #                         CURSES UTILITY
    ###########################################################################

    def basic_screen_setup(self, halfdelay: bool) -> None:
        """
        Basic configurations of the current curses terminal screen

        Args:
            halfdely (bool): If True, refresh specified 1/10th of second, else refresh 25.5 seconds
        Returns:
            None
        """
        # Hiding the cursor
        curses.curs_set(0)

        # Load curses colors
        self.color, self.decor = mu.load_curses_colors_decor()

        # Turn off echo
        curses.noecho()
        curses.nonl()

        # Screen delay/block for 10th of a second
        if halfdelay:
            curses.halfdelay(self.halfdelay_screen_refresh)
        else:
            curses.halfdelay(255)

    def check_terminal_size(self, scr) -> None:
        """
        Checking if current terminal size is sufficient, if it is not, display warning.

        Args:
            scr (obj): Handle for curses terminal screen handle
        Returns:
            None
        """
        # Getting the screen height and width
        term_height, term_width = scr.getmaxyx()

        # Check Height and width
        self.terminal_size_good = term_height >= self.height_limit and term_width >= self.width_limit

        # Debug Terminal Size
        if logger.level == logging.DEBUG:
            scr.addstr(1, 1, f"[Terminal Size: W:{term_width}, H:{term_height}]", self.color['grey-light'])

        k = 0
        ui_keys = mu.load_keys()
        while not self.terminal_size_good:
            if k in ui_keys['QUIT']:
                self.all_threads_enabled = False
                sys.exit(0)

            # Re-evaluate the screen size
            term_height, term_width = scr.getmaxyx()
            self.terminal_size_good = term_height >= self.height_limit and term_width >= self.width_limit

            message_lines = [
                '', 'Uh-Oh! Window too small!', '', '乁( ◔ ౪◔)「    (ಥ⌣ಥ)', '',
                f'Current: W:{term_width} x H:{term_height}', f'Needed : W:{self.width_limit} x H:{self.height_limit}',
                '', 'To continue, resize window', ''
                'To quit, press "Q" or "ESC"'
            ]
            try:
                for y_2, line in enumerate(message_lines):
                    scr.addstr(1 + y_2, mu.get_center_x(scr, line), line, self.decor['bold'])
            except Exception:
                logger.debug(
                    f'Failed to render window. Window size way too small. Needed : W:{self.width_limit} x H:{self.height_limit}'
                )
                self.all_threads_enabled = False
                sys.exit(1)

            scr.refresh()
            k = scr.getch()

    def status_to_color(self, status_text: str) -> str:
        """
        Given a status text, get the color for the status

        Args:
            scr (obj): Handle for curses terminal screen handle
        Returns:
            None
        """
        for status_item in Status:
            if status_text.strip().upper() in status_item.value:
                return Color.ITEMS.value[status_item.name]
        return Color.ITEMS.value['UNKNOWN']

    def status_to_sound(self, status_text: str) -> str:
        """
        Given a status sound, get the sound for the status

        Args:
            scr (obj): Handle for curses terminal screen handle

        Returns:
            None
        """
        for status_item in Status:
            if status_text.strip().upper() in status_item.value:
                return Sound.ITEMS.value[status_item.name]
        return Sound.ITEMS.value['UNKNOWN']

    ###########################################################################
    #                         PLAY SOUND EFFECT
    ###########################################################################

    def __thread_play_sound(self, sound_filepath: str) -> None:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Thread starting - Play sound - (ID: {threading.get_ident()} - Sound: {sound_filepath}s) ...')

        # Load the file and play it
        if platform.system() != "Windows":
            try:
                wave_obj = simpleaudio.WaveObject.from_wave_file(sound_filepath)
                play_obj = wave_obj.play()
            except Exception as error:
                logger.error(f'Failed to play sound. Exception: {error}')
            self.playing_sound = True
            while self.all_threads_enabled:
                if play_obj.is_playing():
                    sleep(0.100)
                    if not self.all_threads_enabled:
                        break
                else:
                    break
        else:
            try:
                winsound.PlaySound(sound_filepath, winsound.SND_FILENAME | winsound.SND_NODEFAULT)
                self.playing_sound = True
            except RuntimeError as error:
                logger.error(f'Failed to play sound. Exception: {error}')
        self.playing_sound = False
        logger.debug(f'Thread stopped - Play sound - (ID: {threading.get_ident()})')

    def play_sound_thread_on(self, sound_filepath: str) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Playing sound file "{sound_filepath}" ...')
        try:
            threading.Thread(target=self.__thread_play_sound, args=(sound_filepath, ), daemon=True).start()
        except Exception as error:
            logger.error(
                f'Failed to start play sound thread for "{sound_filepath}". Exception: {error}. Type: {type(error)}')

        return True

    ###########################################################################
    #                         SERVER STATUS
    ###########################################################################

    def __thread_server_status(self, monitor_interval: float) -> None:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(
            f'Thread starting - Server Status - (ID: {threading.get_ident()} - Refresh Interval: {monitor_interval}s) ...'
        )

        # Set the monitoring thread flag up
        self.all_threads_enabled = True
        self.server_status_thread_interval = monitor_interval

        # Loop until flags disable it
        while self.all_threads_enabled:
            if not self.paused:
                # Get the build information
                self.server_interaction = True
                self.server_status_data['reachable'] = self.rest.is_reachable()
                self.server_status_data['auth'] = self.auth.verify()

            # Wait some time before checking again
            start_time = time()
            while time() - start_time <= monitor_interval:
                sleep(0.1)
                if not self.all_threads_enabled:
                    break

        logger.debug(f'Thread stopped - Server Status - (ID: {threading.get_ident()})')

    def server_status_thread_on(self, monitor_interval: float = 10.0) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug(f'Starting thread for server status ...')
        try:
            threading.Thread(target=self.__thread_server_status, args=(monitor_interval, ), daemon=False).start()
        except Exception as error:
            logger.error(f'Failed to start server status monitoring thread. Exception: {error}. Type: {type(error)}')

        return True

    ###########################################################################
    #                         ALL THREAD CONTROL
    ###########################################################################

    def all_threads_off(self) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # logger.debug(f'Stopping all monitor threads ...')

        # Set the monitoring thread flag down
        self.all_threads_enabled = False

        return True

    def all_threads_pause(self) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        logger.debug('Pausing all monitor threads ...')

        # Set the monitoring thread flag down
        self.paused = True

        return True



"""Stage Menu CLI Entrypoints"""

import logging

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history
from yojenkins.utility.utility import fail_out, print2
from yojenkins.yo_jenkins.status import Status

# Getting the logger reference
logger = logging.getLogger()


@log_to_history
def info(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, stage_name: str, job: str,
         build_number: int, build_url: str, latest: bool) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    if job and not build_number and not latest:
        fail_out('INPUT ERROR: For job, either specify --number or --latest. See --help')

    yj_obj = cu.config_yo_jenkins(profile)

    # Differentiate if name or url
    if build_url:
        logger.debug(f'Build URL passed: {build_url}')
        valid_url_format = True
    else:
        valid_url_format = cu.is_full_url(job) if not build_url else True

    # Request the data
    if valid_url_format:
        data = yj_obj.stage.info(stage_name=stage_name,
                                 build_url=build_url,
                                 job_url=job,
                                 build_number=build_number,
                                 latest=latest)
    else:
        data = yj_obj.stage.info(stage_name=stage_name,
                                 build_url=build_url,
                                 job_name=job,
                                 build_number=build_number,
                                 latest=latest)
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def status(profile: str, stage_name: str, job: str, build_number: int, build_url: str, latest: bool) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    if job and not build_number and not latest:
        fail_out('INPUT ERROR: For job, either specify --number or --latest. See --help')

    yj_obj = cu.config_yo_jenkins(profile)

    # Differentiate if name or url
    if build_url:
        logger.debug(f'Build URL passed: {build_url}')
        valid_url_format = True
    else:
        valid_url_format = cu.is_full_url(job) if not build_url else True

    # Request the data
    if valid_url_format:
        data = yj_obj.stage.status_text(stage_name=stage_name,
                                        build_url=build_url,
                                        job_url=job,
                                        build_number=build_number,
                                        latest=latest)
    else:
        data = yj_obj.stage.status_text(stage_name=stage_name,
                                        build_url=build_url,
                                        job_name=job,
                                        build_number=build_number,
                                        latest=latest)

    # Color for output
    if data.upper() in Status.UNKNOWN.value:
        color = 'black'
    elif data.upper() in Status.QUEUED.value:
        color = 'yellow'
    elif data.upper() in Status.RUNNING.value:
        color = 'blue'
    elif data.upper() in Status.SUCCESS.value:
        color = 'green'
    elif data.upper() in Status.FAILURE.value:
        color = 'red'
    else:
        color = ''
    print2(f'{data}', bold=True, color=color)


@log_to_history
def steps(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, opt_list: bool,
          stage_name: str, job: str, build_number: int, build_url: str, latest: bool) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    if job and not build_number and not latest:
        fail_out('INPUT ERROR: For job, either specify --number or --latest. See --help')

    yj_obj = cu.config_yo_jenkins(profile)

    # Differentiate if name or url
    if build_url:
        logger.debug(f'Build URL passed: {build_url}')
        valid_url_format = True
    else:
        valid_url_format = cu.is_full_url(job) if not build_url else True

    # Request the data
    if valid_url_format:
        data, data_list = yj_obj.stage.step_list(stage_name=stage_name,
                                                 build_url=build_url,
                                                 job_url=job,
                                                 build_number=build_number,
                                                 latest=latest)
    else:
        data, data_list = yj_obj.stage.step_list(stage_name=stage_name,
                                                 build_url=build_url,
                                                 job_name=job,
                                                 build_number=build_number,
                                                 latest=latest)
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def logs(profile: str, stage_name: str, job: str, build_number: int, build_url: str, latest: bool,
         download_dir: bool) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    if job and not build_number and not latest:
        fail_out('INPUT ERROR: For job, either specify --number or --latest. See --help')

    yj_obj = cu.config_yo_jenkins(profile)

    # Differentiate if name or url
    if build_url:
        logger.debug(f'Build URL passed: {build_url}')
        valid_url_format = True
    else:
        valid_url_format = cu.is_full_url(job) if not build_url else True

    # Request the data
    if valid_url_format:
        yj_obj.stage.logs(stage_name=stage_name,
                          build_url=build_url,
                          job_url=job,
                          build_number=build_number,
                          latest=latest,
                          download_dir=download_dir)
    else:
        yj_obj.stage.logs(stage_name=stage_name,
                          build_url=build_url,
                          job_name=job,
                          build_number=build_number,
                          latest=latest,
                          download_dir=download_dir)



"""Account Menu CLI Entrypoints"""

import logging

import click

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history

# Getting the logger reference
logger = logging.getLogger()


@log_to_history
def list(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, opt_list: bool, profile: str) -> None:
    """List all users

    Args:
        opt_pretty: Option to pretty print the output
        opt_yaml: Option to output in YAML format
        opt_xml: Option to output in XML format
        opt_toml: Option to output in TOML format
        opt_list: Option to list the available profiles
        profile: The profile/account to use

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    data, data_list = yj_obj.account.list()
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def info(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, user_id: str) -> None:
    """Get user information

    Args:
        opt_pretty: Option to pretty print the output
        opt_yaml: Option to output in YAML format
        opt_xml: Option to output in XML format
        opt_toml: Option to output in TOML format
        opt_list: Option to list the available profiles
        profile: The profile/account to use
        user_id: The user id to look up

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    data = yj_obj.account.info(user_id=user_id)
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def create(profile: str, user_id: str, password: str, is_admin: bool, email: str, description: str) -> None:
    """Delete a user account

    Args:
        profile: The profile/account to use
        user_id: The user id to look up
        password: The password to set for this user
        is_admin: Is this user an admin
        email: The email address for this user
        description: The description for this user

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    data = yj_obj.account.create(user_id=user_id,
                                 password=password,
                                 is_admin=is_admin,
                                 email=email,
                                 description=description)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def delete(profile: str, user_id: str) -> None:
    """Delete a user account

    Args:
        profile: The profile/account to use
        user_id: The user id to look up

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    data = yj_obj.account.delete(user_id=user_id)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def permission(profile: str, user_id: str, action: str, permission_id: str) -> None:
    """Add or remove user permission

    Args:
        profile: The profile/account to use
        user_id: The user id to look up
        action: The action to perform ("add" or "remove")
        permission_id: The permission to add or remove to action

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    data = yj_obj.account.permission(user_id=user_id, action=action, permission_id=permission_id)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def permission_list(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, opt_list: bool,
                    profile: str) -> None:
    """List all available permissions

    Args:
        opt_pretty: Option to pretty print the output
        opt_yaml: Option to output in YAML format
        opt_xml: Option to output in XML format
        opt_toml: Option to output in TOML format
        opt_list: Option to list the available profiles
        profile: The profile/account to use

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    data, data_list = yj_obj.account.permission_list()
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)



"""Tools Menu CLI Entrypoints"""

import json
import logging
import os
import sys
from pathlib import Path
from typing import Tuple

import click

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history
from yojenkins.tools import Package, SharedLibrary
from yojenkins.utility.utility import (
    browser_open,
    html_clean,
    load_contents_from_local_file,
    print2,
)

# Getting the logger reference
logger = logging.getLogger()

BUG_REPORT_URL = "https://github.com/ismet55555/yojenkins/issues/new?assignees=&labels=&template=bug_report.md&title="
FEATURE_REQUEST_URL = "https://github.com/ismet55555/yojenkins/issues/new?assignees=&labels=&template=feature_request.md&title="
DOCS_URL = "https://www.yojenkins.com/"


@log_to_history
def documentation() -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    logger.debug(f'Opening documentation in web browser: "{DOCS_URL}" ...')
    success = browser_open(url=DOCS_URL)
    if success:
        logger.debug('Successfully opened in web browser')
    else:
        logger.debug('Failed to open in web browser')
    return success


@log_to_history
def upgrade(user: bool, proxy: str) -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    if not Package.install(user=user, proxy=proxy):
        click.secho('failed to upgrade', fg='bright_red', bold=True)
        sys.exit(1)
    click.secho('successfully upgraded', fg='bright_green', bold=True)


@log_to_history
def remove() -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    if click.confirm('Are you sure you want to remove yojenkins?'):
        Package.uninstall()


@log_to_history
def bug_report() -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    logger.debug(f'Opening bug report webpage in web browser: "{BUG_REPORT_URL}" ...')
    success = browser_open(url=BUG_REPORT_URL)
    if success:
        logger.debug('Successfully opened in web browser')
    else:
        logger.debug('Failed to open in web browser')
    return success


@log_to_history
def feature_request() -> None:
    """TODO Docstring

    Details: TODO

    Args:
        TODO

    Returns:
        TODO
    """
    logger.debug(f'Opening feature request webpage in web browser: "{FEATURE_REQUEST_URL}" ...')
    success = browser_open(url=FEATURE_REQUEST_URL)
    if success:
        logger.debug('Successfully opened in web browser')
    else:
        logger.debug('Failed to open in web browser')
    return success


def history(profile: str, clear: bool) -> None:
    """Displaying the command history and clearing the history file if requested.

    # TODO: Ability to clear only for a specific profile.

    Args:
        profile (str): The name of the profile to to filter history with
        clear (bool):  Clearing the history file

    Returns:
        None
    """
    # Load contents from history file
    history_file_path = os.path.join(os.path.join(Path.home(), cu.CONFIG_DIR_NAME), cu.HISTORY_FILE_NAME)
    contents = load_contents_from_local_file('json', history_file_path)
    if not contents:
        click.secho('No history found', fg='bright_red', bold=True)
        sys.exit(1)

    # Clearing the history file if requested
    if clear:
        logger.debug(f'Removing history file: {history_file_path} ...')
        try:
            os.remove(history_file_path)
        except OSError:
            logger.debug('Failed to clear history file')
            click.secho('failed', fg='bright_red', bold=True)
        else:
            logger.debug('Successfully cleared history file')
            click.secho('successfully cleared', fg='bright_green', bold=True)
            sys.exit(0)

    # Displaying the command history
    logger.debug(f'Displaying command history for profile "{profile}" ...')

    def output_history_to_console(command_list: list, profile_name: str) -> None:
        """Helper function to format and output to console"""
        for command_info in command_list:
            profile_str = f'{click.style("[" + profile_name + "]", fg="yellow", bold=True)}'
            datetime_str = f'{click.style("[" + command_info["datetime"] + "]", fg="green", bold=False)}'
            tool_version = f'{click.style("[" + "v" + command_info["tool_version"] + "]", fg="green", bold=False)}'

            command_info = f'{profile_str} {datetime_str} {tool_version} - {command_info["tool_path"]} {command_info["arguments"]}'
            click.echo(command_info)

    if profile:
        if profile in contents:
            output_history_to_console(contents[profile], profile)
        else:
            click.secho('No history found for profile: ' + profile, fg='bright_red', bold=True)
    else:
        for profile_name in contents:
            output_history_to_console(contents[profile_name], profile_name)


@log_to_history
def rest_request(profile: str, request_text: str, request_type: str, raw: bool, clean_html: bool) -> None:
    """Send a generic REST request to Jenkins Server using the loaded credentials

    Args:
        profile (str): The name of the credentials profile
        request_text (str): The text of the request to send
        request_type (str): The type of request to send
        raw (bool): Whether to return the raw response or formatted JSON
        clean_html (bool): Whether to clean the HTML tags from the response

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    request_text = request_text.strip('/')
    content, header, success = yj_obj.rest.request(
        target=request_text,
        request_type=request_type,
        json_content=(not raw),
    )

    if not success:
        click.secho('Failed to make request', fg='bright_red', bold=True)
        sys.exit(1)

    if request_type == 'HEAD':
        print2(header)
        sys.exit(0)

    if content:
        if clean_html:
            try:
                print2(html_clean(content))
            except Exception:
                print2(content)
        else:
            try:
                print2(json.dumps(content, indent=4))
            except Exception:
                print2(content)
    else:
        click.echo(
            click.style('Content returned, however possible HTML content. Try --raw.', fg='bright_red', bold=True))


@log_to_history
def run_script(profile: str, script_text: str, script_filepath: str, output_filepath: str) -> Tuple[str, bool]:
    """TODO

    Details: TODO:

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)

    # Prepare the commands/script
    if script_text:
        script_text = script_text.strip().replace('  ', ' ')
        script = script_text
    elif script_filepath:
        logger.debug(f'Loading specified script from file: {script_filepath} ...')
        try:
            with open(os.path.join(script_filepath), 'r') as open_file:
                script = open_file.read()
            script_size = os.path.getsize(script_filepath)
            logger.debug(f'Successfully loaded script file ({script_size} Bytes)')
        except FileNotFoundError as error:
            click.echo(
                click.style(f'Failed to find specified script file ({script_filepath})', fg='bright_red', bold=True))
            sys.exit(1)

    # Send the request to the server
    content, _, success = yj_obj.rest.request(target='scriptText',
                                              request_type='post',
                                              data={'script': script},
                                              json_content=False)

    if not success:
        click.secho('Failed to make script run request', fg='bright_red', bold=True)
        sys.exit(1)

    # Save script result to file
    if output_filepath:
        logger.debug(f'Saving script result into file: {output_filepath} ...')
        try:
            with open(output_filepath, 'w+') as file:
                file.write(content)
            logger.debug('Successfully wrote script result to file')
        except Exception as error:
            logger.debug(f'Failed to write script result to file. Exception: {error}')
            return "", False

    click.echo(content)


@log_to_history
def shared_lib_setup(profile: str, lib_name: str, repo_owner: str, repo_name: str, repo_url: str, repo_branch: str,
                     implicit: bool, credential_id: str) -> bool:
    """Sets up a shared library on the Jenkins Server

    Args:
        profile: The name of the credentials profile
        lib_name: The name of the shared library
        repo_owner: The owner of the git repo
        repo_name: The name of the git repo
        repo_url: The full url of the git repo (https://<path-to-the-shared-library-repository>)
        repo_branch: The branch of the git repo (ie. master)
        implicit: Whether to use the implicit shared library setup
        credential_id: The Jenkins credential ID to use for the shared library

    Returns:
        True if the setup was successful, else False
    """
    yj_obj = cu.config_yo_jenkins(profile)
    data = SharedLibrary().setup(yj_obj.rest, lib_name, repo_owner, repo_name, repo_url, repo_branch, implicit,
                                 credential_id)
    if not data:
        click.secho('failed', fg='bright_red', bold=True)
        sys.exit(1)
    click.secho('success', fg='bright_green', bold=True)



"""Jenkins item configuration definition"""

from enum import Enum


class JenkinsItemConfig(Enum):
    """Enum of Jenkins item configs ordered in categories for creating

    Usage Examples:
        - `JenkinsItemConfig.FOLDER.value['blank']`
    """

    FOLDER = {
        "blank":
            '''<?xml version='1.1' encoding='UTF-8'?>
                    <com.cloudbees.hudson.plugins.folder.Folder plugin="cloudbees-folder@6.15">
                    <properties/>
                    <folderViews class="com.cloudbees.hudson.plugins.folder.views.DefaultFolderViewHolder">
                        <views>
                        <hudson.model.AllView>
                            <owner class="com.cloudbees.hudson.plugins.folder.Folder" reference="../../../.."/>
                            <name>All</name>
                            <filterExecutors>false</filterExecutors>
                            <filterQueue>false</filterQueue>
                            <properties class="hudson.model.View$PropertyList"/>
                        </hudson.model.AllView>
                        </views>
                        <tabBar class="hudson.views.DefaultViewsTabBar"/>
                    </folderViews>
                    <healthMetrics/>
                    <icon class="com.cloudbees.hudson.plugins.folder.icons.StockFolderIcon"/>
                    </com.cloudbees.hudson.plugins.folder.Folder>''',
        "template":
            ""
    }
    VIEW = {
        "blank":
            '''<?xml version="1.1" encoding="UTF-8"?>
                    <hudson.model.ListView>
                        <name>blankio</name>
                        <filterExecutors>false</filterExecutors>
                        <filterQueue>false</filterQueue>
                        <properties class="hudson.model.View$PropertyList"/>
                        <jobNames>
                            <comparator class="hudson.util.CaseInsensitiveComparator"/>
                        </jobNames>
                        <jobFilters/>
                        <columns>
                            <hudson.views.StatusColumn/>
                            <hudson.views.WeatherColumn/>
                            <hudson.views.JobColumn/>
                            <hudson.views.LastSuccessColumn/>
                            <hudson.views.LastFailureColumn/>
                            <hudson.views.LastDurationColumn/>
                            <hudson.views.BuildButtonColumn/>
                        </columns>
                        <recurse>false</recurse>
                    </hudson.model.ListView>''',
        "template":
            ""
    }
    JOB = {
        "blank":
            '''<?xml version='1.1' encoding='UTF-8'?>
                    <project>
                        <description></description>
                        <keepDependencies>false</keepDependencies>
                        <properties/>
                        <scm class="hudson.scm.NullSCM"/>
                        <canRoam>true</canRoam>
                        <disabled>false</disabled>
                        <blockBuildWhenDownstreamBuilding>false</blockBuildWhenDownstreamBuilding>
                        <blockBuildWhenUpstreamBuilding>false</blockBuildWhenUpstreamBuilding>
                        <triggers/>
                        <concurrentBuild>false</concurrentBuild>
                        <builders/>
                        <publishers/>
                        <buildWrappers/>
                    </project>''',
        "script":
            '''<?xml version='1.1' encoding='UTF-8'?>
                    <project>
                        <actions/>
                        <description></description>
                        <keepDependencies>false</keepDependencies>
                        <properties/>
                        <scm class="hudson.scm.NullSCM"/>
                        <canRoam>true</canRoam>
                        <disabled>false</disabled>
                        <blockBuildWhenDownstreamBuilding>false</blockBuildWhenDownstreamBuilding>
                        <blockBuildWhenUpstreamBuilding>false</blockBuildWhenUpstreamBuilding>
                        <triggers/>
                        <concurrentBuild>false</concurrentBuild>
                        <builders>
                            <hudson.tasks.Shell>
                            <command>
                                {{ SCRIPT }}
                            </command>
                            <configuredLocalRules/>
                            </hudson.tasks.Shell>
                        </builders>
                        <publishers/>
                        <buildWrappers/>
                    </project>'''
    }



"""Build class definition"""

import logging
import os
from datetime import datetime, timedelta
from itertools import islice
from time import sleep
from typing import Dict, List, Tuple

import requests

from yojenkins.monitor import BuildMonitor
from yojenkins.utility import utility
from yojenkins.utility.utility import fail_out, print2
from yojenkins.yo_jenkins.jenkins_item_classes import JenkinsItemClasses
from yojenkins.yo_jenkins.status import BuildStatus

# Getting the logger reference
logger = logging.getLogger()


class Build():
    """Buld class"""

    def __init__(self, rest, auth) -> None:
        """Object constructor method, called at object creation

        Args:
            rest: Rest object
            auth: Auth object

        Returns:
            None
        """
        self.rest = rest
        self.auth = auth
        self.build_monitor = BuildMonitor(rest, auth, self)

        self.build_logs_extension = ".log"

    def info(self,
             build_url: str = '',
             job_name: str = '',
             job_url: str = '',
             build_number: int = None,
             latest: bool = False) -> Dict:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if build_url:
            # Making a direct request using the passed url
            request_url = f"{build_url.strip('/')}/api/json"
            build_info = self.rest.request(request_url, 'get', is_endpoint=False)[0]
            if not build_info:
                fail_out('Failed to get build info')
        else:
            if not job_name and not job_url:
                fail_out('No job name, job url, and build url provided')

            if job_name and not job_url:
                job_url = utility.name_to_url(self.rest.get_server_url(), job_name)

            job_info, _, success = self.rest.request(f'{job_url.strip("/")}/api/json', 'get', is_endpoint=False)
            if not success:
                fail_out(f'Failed getting build info, because failed to request job info: {job_url}')

            # Check if found item type/class is a build
            if job_info['_class'] not in JenkinsItemClasses.JOB.value['class_type']:
                fail_out(f'Failed to match job type/class. The found item is "{job_info["_class"]}"')

            job_last_build_number = job_info['lastBuild']['number'] if 'lastBuild' in job_info else 0

            # If build number is not passed, get the latest build number for job
            if not build_number and latest:
                # Build number not passed and latest flag is not set
                logger.debug(
                    f'No build number passed BUT --latest flag set. Using latest build number for this job: {job_last_build_number}'
                )
                build_number = job_last_build_number
            elif not build_number and not latest:
                fail_out('Failed to specify build. No build number passed and --latest flag not set')
            else:
                # Build number is passed
                if build_number > job_last_build_number:
                    fail_out('Failed to specify build. Build number exceeds last build number for this job')

            logger.debug(f'Getting build info for job "{job_info["fullName"]}, build {build_number} ...')
            build_info, _, success = self.rest.request(f'{job_url.strip("/")}/{build_number}/api/json',
                                                       'get',
                                                       is_endpoint=False)
            if not success:
                fail_out('Failed to request build info')

        # Check if found item type/class is a build
        if build_info['_class'] not in JenkinsItemClasses.BUILD.value['class_type']:
            fail_out(f'Build found, but failed to match build type/class. This item is "{build_info["_class"]}"')

        # Add additional derived information
        if 'timestamp' in build_info:
            build_info['startDatetime'] = datetime.fromtimestamp(build_info['timestamp'] /
                                                                 1000.0).strftime("%A, %B %d, %Y %I:%M:%S")
            build_info['estimatedDurationFormatted'] = str(timedelta(
                seconds=build_info["estimatedDuration"] /
                1000.0))[:-3] if build_info["estimatedDuration"] > 0 else None

            # Check if results are in
            if 'result' in build_info:
                if build_info['result']:
                    build_info['resultText'] = build_info['result']
                    build_info['durationFormatted'] = str(timedelta(seconds=build_info['duration'] / 1000.0))[:-3]
                    build_info['endDatetime'] = datetime.fromtimestamp(
                        (build_info['timestamp'] + build_info['duration']) / 1000.0).strftime("%A, %B %d, %Y %I:%M:%S")
                    build_info['elapsedFormatted'] = build_info['durationFormatted']
                else:
                    build_info['resultText'] = BuildStatus.RUNNING.value
                    build_info['durationFormatted'] = None
                    build_info['endDatetime'] = None
                    build_info['elapsedFormatted'] = str(
                        timedelta(seconds=((datetime.utcnow().timestamp()) - build_info['timestamp'] / 1000)))[:-3]

            else:
                build_info['resultText'] = BuildStatus.UNKNOWN.value
        else:
            build_info['startDatetime'] = None
            build_info['estimatedDurationFormatted'] = None
            build_info['resultText'] = BuildStatus.NOT_RUN.value
            build_info['durationFormatted'] = None
            build_info['endDatetime'] = None
            build_info['elapsedFormatted'] = None

        if 'url' in build_info:
            build_info['fullName'] = utility.url_to_name(build_info['url'])
            build_info['jobUrl'] = utility.build_url_to_other_url(build_info['url'], target_url='job')
            build_info['jobFullName'] = utility.url_to_name(build_info['jobUrl'])
            build_info['jobName'] = utility.fullname_to_name(build_info['jobFullName'])
            build_info['folderUrl'] = utility.build_url_to_other_url(build_info['url'], target_url='folder')
            build_info['folderFullName'] = utility.url_to_name(build_info['folderUrl'])
            build_info['folderName'] = utility.fullname_to_name(build_info['folderFullName'])
            build_info['serverURL'] = utility.item_url_to_server_url(build_info['url'])
            build_info['serverDomain'] = utility.item_url_to_server_url(build_info['url'], False)

            build_info[
                'folderFullName'] = 'Base Folder' if not build_info['folderFullName'] else build_info['folderFullName']

        if 'builtOn' not in build_info:
            build_info['builtOn'] = 'N/A'
        else:
            build_info['builtOn'] = 'N/A' if not build_info['builtOn'] else build_info['builtOn']

        return build_info

    def status_text(self,
                    build_url: str = '',
                    job_name: str = '',
                    job_url: str = '',
                    build_number: int = None,
                    latest: bool = False) -> str:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Get the build info
        build_info = self.info(build_url=build_url,
                               job_name=job_name,
                               job_url=job_url,
                               build_number=build_number,
                               latest=latest)

        # If nothing is returned, check if job is queued on server
        logger.debug('The specified build was not found in job')
        logger.debug('Looking for build in the server build queue ...')
        if build_url:
            job_url = utility.build_url_to_other_url(build_url)
        elif job_name:
            pass
        elif job_url:
            pass
        else:
            fail_out('Failed to find build status text. Specify build url, job name, or job url')
        logger.debug(f'Job name: {job_name}')

        # Requesting all queue and searching queue (NOTE: Could use Server object)
        logger.debug(f'Requesting all build queue items ...')
        queue_all = self.rest.request('queue/api/json', 'get')[0]
        logger.debug(f"Number of queued items found: {len(queue_all['items'])}")
        queue_matches = utility.queue_find(queue_all, job_name=job_name, job_url=job_url)
        if not queue_matches:
            fail_out('Failed to find running or queued builds')
        queue_info = queue_matches[0]

        if not queue_info:
            logger.debug('Build for job NOT found in queue')
            return BuildStatus.NOT_FOUND.value
        else:
            logger.debug(f'Build for job found in queue. Queue number {queue_info["id"]}')
            return BuildStatus.QUEUED.value

        # FIXME: resultText is returned in build info. Maybe move queue check to build_info??

        # Check if in process (build is there but results not posted)
        if 'result' not in build_info:
            logger.debug('Build was found running/building, however no results are posted')
            return BuildStatus.RUNNING.value
        else:
            # FIXME: Get "No status found" when "yojenkins build status --url" on build that is "RUNNING" (result: Null)
            logger.debug('Build found, but has concluded or stopped with result')
            return build_info['result']

    def abort(self,
              build_url: str = '',
              job_name: str = '',
              job_url: str = '',
              build_number: int = None,
              latest: bool = False) -> int:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            url = build_url
        else:
            logger.debug('No build URL passed. Getting build information ...')
            # Get build info request
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            url = build_info['url']

        # Making a direct request using the passed url
        logger.debug(f'Aborting build: {url} ...')
        request_url = f"{url.strip('/')}/stop"
        if not self.rest.request(request_url, 'post', is_endpoint=False)[2]:
            fail_out('Failed to abort build. Build may not exist or is queued')

        logger.debug('Successfully aborted build')

        return utility.build_url_to_build_number(build_url=url)

    def delete(self,
               build_url: str = '',
               job_name: str = '',
               job_url: str = '',
               build_number: int = None,
               latest: bool = False) -> int:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO: Pass a list of build numbers
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            url = build_url
        else:
            logger.debug('No build URL passed. Getting build information ...')
            # Get build info request
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            url = build_info['url']

        # Making a direct request using the passed url
        logger.debug(f'Deleting build: {url} ...')
        request_url = f"{url.strip('/')}/doDelete"
        if not self.rest.request(request_url, 'post', is_endpoint=False)[2]:
            fail_out('Failed to delete build. Build may not exist or is queued')

        return utility.build_url_to_build_number(build_url=url)

    def stage_list(self,
                   build_url: str = '',
                   job_name: str = '',
                   job_url: str = '',
                   build_number: int = None,
                   latest: bool = False) -> Tuple[list, list]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # FIXME: yojenkins build stages --url https://localhost:8080/job/Non-PAR/job/Non-Prod-Jobs/job/Something/job/test_job/46/
        #        yields 404 in running build. Maybe issue with formatting of the url to name?

        # TODO: Pass a list of build numbers
        if not build_url:
            logger.debug('No build URL passed. Getting build information ...')
            # Get build info request
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            build_url = build_info['url']

        # Making a direct request using the passed url
        logger.debug(f'Getting build stages for: {build_url} ...')
        request_url = f"{build_url.strip('/')}/wfapi/describe"
        return_content, _, return_success = self.rest.request(request_url, 'get', is_endpoint=False)
        if not return_success or not return_content:
            fail_out('Failed to get build stages. This may not be a staged build')

        # Getting the stage items
        # FIXME: When --url <job> and no build number is passed, it will just get the job describe, not build info
        if 'stages' in return_content:
            build_stage_list = return_content['stages']
        else:
            fail_out('Failed to find "stages" key in build info. This may not be a staged build')

        # Add additional derived information for each step
        for stage_info in build_stage_list:
            stage_info['startDatetime'] = datetime.fromtimestamp(stage_info["startTimeMillis"] /
                                                                 1000.0).strftime("%A, %B %d, %Y %I:%M:%S")
            stage_info['durationFormatted'] = str(timedelta(seconds=stage_info["durationMillis"] / 1000.0))[:-3]
            stage_info['pauseDurationFormatted'] = str(timedelta(seconds=stage_info["pauseDurationMillis"] / 1000.0))
            stage_info['url'] = stage_info['_links']['self']['href']

        # Getting only the names of the stages
        build_stage_name_list = [stage['name'] for stage in build_stage_list]

        return build_stage_list, build_stage_name_list

    def artifact_list(self,
                      build_url: str = '',
                      job_name: str = '',
                      job_url: str = '',
                      build_number: int = None) -> List[dict]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Test on build with artifacts
        return self.info(build_url=build_url, job_name=job_name, job_url=job_url,
                         build_number=build_number).get('artifacts')

    def artifact_download(self):
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO: Test on build with artifacts
        pass

    def logs(self,
             build_url: str = '',
             job_name: str = '',
             job_url: str = '',
             build_number: int = None,
             latest: bool = False,
             tail: float = None,
             download_dir: str = '',
             follow: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            url = build_url
        else:
            logger.debug('No build URL passed. Getting build information through job ...')
            # Get build info request
            build_info = self.info(job_name=job_name, job_url=job_url, build_number=build_number, latest=latest)
            url = build_info['url']

        # FIXME: Check if this is an actual build

        request_url = f"{url.strip('/')}/consoleText"

        if download_dir:
            # Download to local file
            auth = requests.auth.HTTPBasicAuth(self.rest.username, self.rest.api_token)
            filename = f'build-logs_{datetime.now().strftime("%m-%d-%Y_%I-%M-%S")}{self.build_logs_extension}'
            logger.debug(f'Downloading console text logs to local file "{filename}" ...')
            try:
                with requests.get(request_url, auth=auth, stream=True, timeout=10) as open_request:
                    open_request.raise_for_status()
                    with open(os.path.join(download_dir, filename), 'wb') as open_file:
                        for chunk in open_request.iter_content(chunk_size=8192):
                            if chunk:
                                open_file.write(chunk)
                logger.debug('Successfully download build logs to file')
            except Exception as error:
                fail_out(f'Failed to download or save logs for build. Exception: {error}')
        else:
            # Stream the logs to console
            if not follow:
                logger.debug('Fetching logs from server ...')
                return_content, _, return_success = self.rest.request(request_url,
                                                                      'get',
                                                                      is_endpoint=False,
                                                                      json_content=False)
                if not return_success or not return_content:
                    fail_out('Failed to get console logs. Build may not exist or is queued')

                # If tail/last part of the log was specified
                if tail:
                    logger.debug(f'--tail option specified with value of: {tail}')
                    tail = abs(tail)
                    logs_list = list(map(lambda num: num.strip(), return_content.splitlines()))
                    number_of_lines = round(len(logs_list) * tail) if tail < 1 else round(tail)
                    start_log_number = 0 if number_of_lines > len(logs_list) else len(logs_list) - number_of_lines
                    return_content = os.linesep.join(list(islice(logs_list, start_log_number, None)))
                    logger.debug(f'Only printing out last logs, lines {start_log_number} to {len(logs_list)} ...')

                logger.debug('Printing out console text logs ...')
                print2(return_content)
            else:
                logger.debug('Following/streaming logs from server ...')
                logger.debug(
                    'NOTE: Jenkins server does not support requesting partial byte ranges, MUST download entire log to get log message differences'
                )
                auth = requests.auth.HTTPBasicAuth(self.rest.username, self.rest.api_token)
                old_dict = {}
                fetch_number = 1
                try:
                    while True:
                        # FIXME: Preserve already used session!  Use Rest object
                        response = requests.head(request_url, auth=auth)
                        if 'content-length' not in response.headers:
                            fail_out(
                                f'Failed to find "content-length" key in server response headers: {response.headers}')
                        content_length_sample_1 = int(response.headers['Content-Length'])
                        sleep(1)
                        response = requests.head(request_url, auth=auth)
                        content_length_sample_2 = int(response.headers['Content-Length'])

                        content_length_diff = content_length_sample_2 - content_length_sample_1
                        if content_length_diff:
                            logger.debug(
                                f'FETCH {fetch_number}: Content length difference: {content_length_diff} bytes')
                            return_content, _, return_success = self.rest.request(request_url,
                                                                                  'get',
                                                                                  is_endpoint=False,
                                                                                  json_content=False)
                            # new_dict = dict.fromkeys([ x for x in (line.strip() for line in return_content.splitlines()) if x ])
                            new_dict = dict.fromkeys(list(map(lambda num: num.strip(), return_content.splitlines())))

                            diff = dict.fromkeys(x for x in new_dict if x not in old_dict)
                            diff_list = list(diff.keys())
                            diff_text = os.linesep.join(diff_list)

                            old_dict = new_dict
                            fetch_number += 1

                            print2(diff_text)
                        else:
                            logger.debug('No content length difference')
                except KeyboardInterrupt:
                    logger.debug('Keyboard Interrupt (CTRL-C) by user. Stopping log following ...')
        return True

    def browser_open(self,
                     build_url: str = '',
                     job_name: str = '',
                     job_url: str = '',
                     build_number: int = None,
                     latest: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # Need URL
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            build_url = build_url.strip('/')
        else:
            logger.debug('No build URL passed. Getting build information ...')
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            build_url = build_info['url']

        # Open the build in browser
        logger.debug(f'Opening build in web browser: "{build_url}" ...')
        success = utility.browser_open(url=build_url)
        if not success:
            fail_out('Failed to open build in web browser')
        logger.debug('Successfully opened build in web browser')

        return success

    def monitor(self,
                build_url: str = '',
                job_name: str = '',
                job_url: str = '',
                build_number: int = None,
                latest: bool = False,
                sound: bool = False) -> bool:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        if build_url:
            logger.debug(f'Build URL passed: {build_url}')
            if not self.rest.request(f"{build_url.strip('/')}/api/json", 'head', is_endpoint=False)[0]:
                fail_out(f'Failed to find build. The build may not exist: {build_url}')
            url = build_url
        else:
            logger.debug('No build URL passed. Getting build information ...')
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            url = build_info['url']

        logger.debug(f'Starting monitor for: "{url}" ...')
        success = self.build_monitor.monitor_start(build_url=url, sound=sound)
        if not success:
            fail_out('Failed to start build monitor')
        logger.debug('Successfully started build monitor')

        return success

    def parameters(self,
                   build_url: str = '',
                   job_name: str = '',
                   job_url: str = '',
                   build_number: int = None,
                   latest: bool = False) -> Tuple[list, list]:
        """TODO Docstring

        Args:
            TODO

        Returns:
            TODO
        """
        # TODO: Pass a list of build numbers
        if not build_url:
            logger.debug('No build URL passed. Getting build information ...')
            # Get build info request
            build_info = self.info(build_url, job_name, job_url, build_number, latest)
            build_url = build_info['url']

        logger.debug(f'Getting build parameters for: "{build_url}" ...')

        # Check if item has any parameter actions
        parameter_actions = utility.get_item_action(build_info, 'hudson.model.ParametersAction')
        if not parameter_actions:
            fail_out('This build does not have any build parameters')

        # Get the parameter definitions
        parameters = parameter_actions[0]['parameters']

        # List of parameter items
        parameters_list = [
            f'{parameter["_class"].split(".")[-1]} - {parameter["name"]} - {parameter["value"]}'
            for parameter in parameters
        ]

        return parameters, parameters_list



"""Installation/Packaging Definition for Python Project"""

import os
import sys

import setuptools

# Package version number (Updated via bump2version tool)
__version__ = "0.0.57"

def check_py_version():
    """Check the python version"""
    if sys.version_info < (3, 7):
        print('Your Python version ({}.{}) is not supported'.format(sys.version_info.major, sys.version_info.minor))
        print('You must have Python version 3.7 or higher to run this program')
        if sys.version_info.major < 3:
            print('You may have used "python" where you needed to use "python3"?')
        sys.exit(1)

def get_requirements():
    """Load packages from requirements.txt"""

    check_py_version()

    with open(os.path.join(os.path.dirname(__file__), "requirements.txt")) as handle:
        packages = handle.readlines()
    packages = [package.strip() for package in packages]

    return packages

def get_pipfile_requirements():
    """Load packages from Pipfile"""
    from toml import loads

    # Loading Pipfile
    try:
        with open ('Pipfile', 'r') as open_file:
            pipfile = open_file.read()
        pipfile_toml = loads(pipfile)
    except FileNotFoundError:
        return []

    # Getting the "packages section"
    try:
        required_packages = pipfile_toml['packages'].items()
    except KeyError:
        return []

    # Parsing Pipfile
    packages = []
    for package_name, ver in required_packages:
        version = ver
        if not isinstance(ver, str):  # Check for a dict
            version_parts = []
            if 'version' in ver:
                version_parts.append(ver['version'] if ver['version'] != '*' else "")
            if 'platform_system' in ver:
                version_parts.append(f"platform_system {ver['platform_system']}")
            version = '; '.join(version_parts)
        packages.append(package_name if version == "*" else f"{package_name}{version}")
    return packages

def read(fname):
    """Utility function to read the README file. Used for the long_description"""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setuptools.setup(
    name="yojenkins",
    version=__version__,
    author="Ismet Handzic",
    author_email="ismet.handzic@gmail.com",
    maintainer="Ismet Handzic",
    description="A CLI tool to manage and have fun with Jenkins server",
    keywords="jenkins monitor manage job build fun",
    url="https://github.com/ismet55555/yojenkins",
    packages=setuptools.find_packages(),
    install_requires=get_requirements(),
    include_package_data=True,
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    python_requires='>=3.7',
    setup_requires=['wheel'],
    py_modules=["yojenkins"],
    entry_points={
        "console_scripts": [
                "yojenkins = yojenkins.__main__:main"
            ]
        },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'Topic :: Software Development :: Testing',
        'Topic :: System :: Monitoring',
        'Topic :: Software Development :: Libraries :: Python Modules',
        "Environment :: Console",
        "Environment :: Console :: Curses",
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only'
    ]
)



"""YoJenkins class definition"""

import logging

from yojenkins.yo_jenkins.account import Account
from yojenkins.yo_jenkins.build import Build
from yojenkins.yo_jenkins.credential import Credential
from yojenkins.yo_jenkins.folder import Folder
from yojenkins.yo_jenkins.job import Job
from yojenkins.yo_jenkins.node import Node
from yojenkins.yo_jenkins.server import Server
from yojenkins.yo_jenkins.stage import Stage
from yojenkins.yo_jenkins.step import Step

# Getting the logger reference
logger = logging.getLogger()


class YoJenkins:
    """This class is a composite class of other objects

    All work is done by other, included, objects
    """

    def __init__(self, auth: object) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.auth = auth
        self.rest = self.auth.get_rest()
        self.jenkins_sdk = self.auth.jenkins_sdk
        self.server = Server(self.rest, self.auth)
        self.node = Node(self.rest)
        self.account = Account(self.rest)
        self.credential = Credential(self.rest)
        self.folder = Folder(self.rest, self.jenkins_sdk)
        self.build = Build(self.rest, self.auth)
        self.job = Job(self.rest, self.folder, self.jenkins_sdk, self.auth, self.build)
        self.step = Step(self.rest)
        self.stage = Stage(self.rest, self.build, self.step)



#!/usr/bin/env python3

import json
import logging
import sys

import click
import xmltodict

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history
from yojenkins.utility.utility import print2

# Getting the logger reference
logger = logging.getLogger()


@log_to_history
def info(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, folder: str) -> None:
    """Folder information

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data = yj_obj.folder.info(folder_url=folder)
    else:
        data = yj_obj.folder.info(folder_name=folder)
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def search(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, search_pattern: str,
           search_folder: str, depth: int, fullname: bool, opt_list: bool) -> None:
    """Search folders by REGEX pattern

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(search_folder):
        data, data_list = yj_obj.folder.search(search_pattern=search_pattern,
                                               folder_url=search_folder,
                                               folder_depth=depth,
                                               fullname=fullname)
    else:
        data, data_list = yj_obj.folder.search(search_pattern=search_pattern,
                                               folder_name=search_folder,
                                               folder_depth=depth,
                                               fullname=fullname)
    if not data:
        print2("No folders found", color="yellow")
        sys.exit(1)
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def subfolders(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, folder: str,
               opt_list: bool) -> None:
    """List all subfolders in folder

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data, data_list = yj_obj.folder.subfolder_list(folder_url=folder)
    else:
        data, data_list = yj_obj.folder.subfolder_list(folder_name=folder)
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def jobs(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, folder: str,
         opt_list: bool) -> None:
    """List all jobs in folder

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data, data_list = yj_obj.folder.jobs_list(folder_url=folder)
    else:
        data, data_list = yj_obj.folder.jobs_list(folder_name=folder)
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def views(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, folder: str,
          opt_list: int) -> None:
    """List all views in folder

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data, data_list = yj_obj.folder.view_list(folder_url=folder)
    else:
        data, data_list = yj_obj.folder.view_list(folder_name=folder)
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def items(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, folder: str,
          opt_list: int) -> None:
    """List all items in folder

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data, data_list = yj_obj.folder.item_list(folder_url=folder)
    else:
        data, data_list = yj_obj.folder.item_list(folder_name=folder)
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def browser(profile: str, folder: str) -> None:
    """Open folder in web browser

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        yj_obj.folder.browser_open(folder_url=folder)
    else:
        yj_obj.folder.browser_open(folder_name=folder)


@log_to_history
def config(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, opt_json: bool, profile: str, folder: str,
           filepath: str) -> None:
    """Get folder configuration

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data = yj_obj.folder.config(filepath=filepath,
                                    folder_url=folder,
                                    opt_json=opt_json,
                                    opt_yaml=opt_yaml,
                                    opt_toml=opt_toml)
    else:
        data = yj_obj.folder.config(filepath=filepath,
                                    folder_name=folder,
                                    opt_json=opt_json,
                                    opt_yaml=opt_yaml,
                                    opt_toml=opt_toml)
    # Converting XML to dict
    # data = json.loads(json.dumps(xmltodict.parse(data)))

    opt_xml = not any([opt_json, opt_yaml, opt_toml])
    data = data if opt_xml else json.loads(json.dumps(xmltodict.parse(data)))
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def create(profile: str, name: str, folder: str, type: str, config: str, config_is_json: bool) -> None:
    """Create an item

    Args:
        TODO

    Returns:
        None
    """
    # TODO: Maybe return the newly created item url

    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data = yj_obj.folder.create(name=name,
                                    type=type,
                                    config=config,
                                    folder_url=folder,
                                    config_is_json=config_is_json)
    else:
        data = yj_obj.folder.create(name=name,
                                    type=type,
                                    config=config,
                                    folder_name=folder,
                                    config_is_json=config_is_json)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def copy(profile: str, folder: str, original_name: str, new_name: str) -> None:
    """Copy an existing item

    Args:
        TODO

    Returns:
        None
    """
    # TODO: Maybe return the newly copied item url

    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data = yj_obj.folder.copy(original_name=original_name, new_name=new_name, folder_url=folder)
    else:
        data = yj_obj.folder.copy(original_name=original_name, new_name=new_name, folder_name=folder)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def delete(profile: str, folder: str) -> None:
    """Delete folder or view

    Args:
        TODO

    Returns:
        None
    """
    yj_obj = cu.config_yo_jenkins(profile)
    if cu.is_full_url(folder):
        data = yj_obj.folder.delete(folder_url=folder)
    else:
        data = yj_obj.folder.delete(folder_name=folder)
    click.secho('success', fg='bright_green', bold=True)



"""Server click sub-command"""
# pylint: skip-file

import click

from yojenkins.__main__ import server
from yojenkins.cli import cli_decorators, cli_server
from yojenkins.cli.cli_utility import set_debug_log_level


@server.command(short_help='\tServer information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
def info(debug, pretty, yaml, toml, xml, profile):
    """Server information"""
    set_debug_log_level(debug)
    cli_server.info(pretty, yaml, xml, toml, profile)


@server.command(short_help='\tShow all people/users on server')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@cli_decorators.list
def people(debug, pretty, yaml, xml, toml, profile, list):
    """Show all people/users on server"""
    set_debug_log_level(debug)
    cli_server.people(pretty, yaml, xml, toml, profile, list)


@server.command(short_help='\tShow current job build queues on server')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@cli_decorators.list
def queue(debug, pretty, yaml, xml, toml, profile, list):
    """Show current job build queues on server"""
    set_debug_log_level(debug)
    cli_server.queue(pretty, yaml, xml, toml, profile, list)
    # NOTE: Maybe move to "job"?


@server.command(short_help='\tShow plugin information')
@cli_decorators.debug
@cli_decorators.format_output
@cli_decorators.profile
@cli_decorators.list
def plugins(debug, pretty, yaml, xml, toml, profile, list):
    """Show plugin information"""
    set_debug_log_level(debug)
    cli_server.plugins(pretty, yaml, xml, toml, profile, list)


@server.command(short_help='\tOpen server home page in web browser')
@cli_decorators.debug
@cli_decorators.profile
def browser(debug, profile):
    """Open server home page in web browser"""
    set_debug_log_level(debug)
    cli_server.browser(profile)


@server.command(short_help='\tCheck if server is reachable')
@cli_decorators.debug
@cli_decorators.profile
@click.option('--timeout', type=int, default=10, required=False, is_flag=False, help='Request timeout value')
def reachable(debug, profile, timeout):
    """Check if server is reachable"""
    set_debug_log_level(debug)
    cli_server.reachable(profile, timeout)


@server.command(short_help='\tServer quite mode enable/disable')
@cli_decorators.debug
@cli_decorators.profile
@click.option('--off', type=bool, default=False, required=False, is_flag=True, help='Undo quiet down mode')
def quiet(debug, profile, off):
    """
    NOTE: A server with quiet mode enabled does not allow any new jobs to be build.
    This may be needed prior to server maintenance, restarts, or shutdowns
    """
    set_debug_log_level(debug)
    cli_server.quiet(profile, off)


@server.command(short_help='\tRestart the server')
@cli_decorators.debug
@cli_decorators.profile
@click.option('--force',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Force restart. Without initial quiet mode.')
def restart(debug, profile, force):
    """
    NOTE: By default this will put Jenkins into the quiet mode, wait for existing builds to be completed, and then restart Jenkins.
    Use --force to skip quiet mode.
    """
    set_debug_log_level(debug)
    cli_server.restart(profile, force)


@server.command(short_help='\tShut down the server')
@cli_decorators.debug
@cli_decorators.profile
@click.option('--force',
              type=bool,
              default=False,
              required=False,
              is_flag=True,
              help='Force shutdown. Without initial quiet mode')
def shutdown(debug, profile, force):
    """
    NOTE: By default this will put Jenkins in a quiet mode, in preparation for a shutdown.
    In that mode Jenkins does not start any new builds.
    Use --force to skip quiet mode.
    """
    set_debug_log_level(debug)
    cli_server.shutdown(profile, force)


@server.command(short_help='\tCreate a local development server using Docker')
@cli_decorators.debug
@click.option('--config-file',
              default='config_as_code.yaml',
              type=click.Path(file_okay=True, dir_okay=False),
              required=False,
              help='.yml/.yaml file for custom configuration as code for Jenkins server setup')
@click.option('--plugins-file',
              default='plugins.txt',
              type=click.Path(file_okay=True, dir_okay=False),
              required=False,
              help='.txt file for custom list of all plugins to be installed on Jenkins server')
@click.option('--protocol-schema',
              default='http',
              type=str,
              required=False,
              help='Protocol schema for Jenkins, http, https, etc.')
@click.option('--host',
              default='localhost',
              type=str,
              required=False,
              help='Jenkins server host (localhost, 192.168.0.1, etc.)')
@click.option('--port', default=8080, type=int, required=False, help='Jenkins server port')
@click.option('--image-base',
              default='jenkins/jenkins',
              show_default=True,
              type=str,
              required=False,
              help='Base Jenkins server image')
@click.option('--image-rebuild',
              default=False,
              type=bool,
              required=False,
              is_flag=True,
              help='If image exists, rebuild existing docker image')
@click.option('--new-volume',
              default=False,
              show_default=True,
              type=bool,
              required=False,
              is_flag=True,
              help='Erase existing Docker data volume from previously created servers')
@click.option('--new-volume-name',
              default='yojenkins-jenkins',
              type=str,
              required=False,
              help='Name of the resulting Docker volume')
@click.option('--bind-mount-dir',
              default='',
              type=click.Path(file_okay=False, dir_okay=True),
              required=False,
              help='Path of local directory to be bound inside container "/tmp/my_things" directory')
@click.option('--container-name',
              default='yojenkins-jenkins',
              type=str,
              required=False,
              help='Name of the resulting Docker container')
@click.option('--registry', default='', type=str, required=False, help='Registry to pull base Jenkins image from')
@click.option('--admin-user',
              default='admin',
              show_default=True,
              type=str,
              required=False,
              help='Set username of admin')
@click.option('--password',
              default='',
              type=str,
              required=False,
              help='Set password for admin account [default: password]')
def server_deploy(debug, config_file, plugins_file, protocol_schema, host, port, image_base, image_rebuild, new_volume,
                  new_volume_name, bind_mount_dir, container_name, registry, admin_user, password):
    """ATTENTION: The resulting Jenkins server is for development and testing purposes only. Enjoy responsibly.

    NOTE: Docker must be installed for this command to function

    BTW: All options have default values and command can be run without any specified options
    """
    set_debug_log_level(debug)
    cli_server.server_deploy(config_file, plugins_file, protocol_schema, host, port, image_base, image_rebuild,
                             new_volume, new_volume_name, bind_mount_dir, container_name, registry, admin_user,
                             password)


@server.command(short_help='\tRemove a local development server')
@click.option('--remove-volume',
              default=False,
              show_default=True,
              type=bool,
              required=False,
              is_flag=True,
              help='Also remove the Docker volume used for current server')
@click.option('--remove-image',
              default=False,
              show_default=True,
              type=bool,
              required=False,
              is_flag=True,
              help='Also remove the Docker image used for current server')
@cli_decorators.debug
def server_teardown(debug, remove_volume, remove_image):
    """Remove a local development server"""
    set_debug_log_level(debug)
    cli_server.server_teardown(remove_volume, remove_image)


# @server.command(short_help='\tCheck if a locally deployed development server is running')
# @cli_decorators.debug
# def server_check(debug):
#     set_debug_log_level(debug)
#     click.secho('TODO :-/', fg='yellow')



"""Build Menu CLI Entrypoints"""

import logging
import sys

import click

from yojenkins.cli import cli_utility as cu
from yojenkins.cli.cli_utility import log_to_history
from yojenkins.yo_jenkins.status import Status

# Getting the logger reference
logger = logging.getLogger()


def _verify_build_url_get_job_format(build_url: str, job: str) -> bool:
    """Utility function to verify buld url and determine if job is URL or name

    Args:
        build_url: Build URL
        job: Job URL or name

    Returns:
         True if both are valid, False otherwise
    """
    if build_url:
        if not cu.is_full_url(build_url):
            click.secho('The build url not formatted correctly', fg='bright_red', bold=True)
            sys.exit(1)
        valid_url_format = True
    else:
        valid_url_format = cu.is_full_url(job)
    return valid_url_format


@log_to_history
def info(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, job: str, build_number: int,
         build_url: str, latest: bool) -> None:
    """TODO Docstring

    Args:
        opt_pretty: Option to pretty print the output
        opt_yaml: Option to output in YAML format
        opt_xml: Option to output in XML format
        opt_toml: Option to output in TOML format
        profile: The profile/account to use
        job: The job this build is under
        build_number: The build number to get info on
        build_url: The build url to get info on
        latest: Option to get the latest build

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        data = yj_obj.build.info(build_url=build_url, job_url=job, build_number=build_number, latest=latest)
    else:
        data = yj_obj.build.info(build_url=build_url, job_name=job, build_number=build_number, latest=latest)
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def status(profile: str, job: str, build_number: int, build_url: str, latest: bool) -> None:
    """Build status text/label

    Args:
        profile: The profile/account to use
        job: The job this build is under
        build_number: The build number to get info on
        build_url: The build url to get info on
        latest: Option to get the latest build

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        data = yj_obj.build.status_text(build_url=build_url, job_url=job, build_number=build_number, latest=latest)
    else:
        data = yj_obj.build.status_text(build_url=build_url, job_name=job, build_number=build_number, latest=latest)

    # Color for output
    if data in Status.NONE.value:
        data = "NONE"
        output_fg = 'black'
    elif data.upper() in Status.UNKNOWN.value:
        output_fg = 'black'
    elif data.upper() in Status.QUEUED.value:
        output_fg = 'yellow'
    elif data.upper() in Status.RUNNING.value:
        output_fg = 'blue'
    elif data.upper() in Status.SUCCESS.value:
        output_fg = 'bright_green'
    elif data.upper() in Status.FAILURE.value:
        output_fg = 'bright_red'
    else:
        output_fg = ''
    click.secho(f'{data}', fg=output_fg, bold=True)


@log_to_history
def abort(profile: str, job: str, build_number: int, build_url: str, latest: bool) -> None:
    """Abort build

    Args:
        profile: The profile/account to use
        job: The job this build is under
        build_number: The build number to get info on
        build_url: The build url to get info on
        latest: Option to get the latest build

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        data = yj_obj.build.abort(build_url=build_url, job_url=job, build_number=build_number, latest=latest)
    else:
        data = yj_obj.build.abort(build_url=build_url, job_name=job, build_number=build_number, latest=latest)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def delete(profile: str, job: str, build_number: int, build_url: str, latest: bool) -> None:
    """Delete build

    Args:
        profile: The profile/account to use
        job: The job this build is under
        build_number: The build number to get info on
        build_url: The build url to get info on
        latest: Option to get the latest build

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        data = yj_obj.build.delete(build_url=build_url, job_url=job, build_number=build_number, latest=latest)
    else:
        data = yj_obj.build.delete(build_url=build_url, job_name=job, build_number=build_number, latest=latest)
    click.secho('success', fg='bright_green', bold=True)


@log_to_history
def stages(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, opt_list: bool, job: str,
           build_number: int, build_url: str, latest: bool) -> None:
    """Get build stages

    Args:
        opt_pretty: Option to pretty print the output
        opt_yaml: Option to output in YAML format
        opt_xml: Option to output in XML format
        opt_toml: Option to output in TOML format
        profile: The profile/account to use
        opt_list: Option to list all stages without details
        job: The job name under which the build is located
        build_number: The build number
        build_url: The build URL
        latest: Option to get the latest build

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        data, data_list = yj_obj.build.stage_list(build_url=build_url,
                                                  job_url=job,
                                                  build_number=build_number,
                                                  latest=latest)
    else:
        data, data_list = yj_obj.build.stage_list(build_url=build_url,
                                                  job_name=job,
                                                  build_number=build_number,
                                                  latest=latest)
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)


@log_to_history
def logs(profile: str, job: str, build_number: int, build_url: str, latest: bool, tail: float, download_dir: bool,
         follow: bool) -> None:
    """Get build logs

    Args:
        profile: The profile/account to use
        job: The job this build is under
        build_number: The build number to get info on
        build_url: The build url to get info on
        latest: Option to get the latest build
        tail: Option to get the last N lines of the log
        download_dir: Option to download the log to a directory
        follow: Option to follow the log

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        yj_obj.build.logs(build_url=build_url,
                          job_url=job,
                          build_number=build_number,
                          latest=latest,
                          tail=tail,
                          download_dir=download_dir,
                          follow=follow)
    else:
        yj_obj.build.logs(build_url=build_url,
                          job_name=job,
                          build_number=build_number,
                          latest=latest,
                          tail=tail,
                          download_dir=download_dir,
                          follow=follow)
    if download_dir:
        click.secho('success', fg='bright_green', bold=True)


@log_to_history
def browser(profile: str, job: str, build_number: int, build_url: str, latest: bool) -> None:
    """Open build in web browser

    Args:
        profile: The profile/account to use
        job: The job this build is under
        build_number: The build number to get info on
        build_url: The build url to get info on
        latest: Option to get the latest build

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        yj_obj.build.browser_open(build_url=build_url, job_url=job, build_number=build_number, latest=latest)
    else:
        yj_obj.build.browser_open(build_url=build_url, job_name=job, build_number=build_number, latest=latest)


@log_to_history
def monitor(profile: str, job: str, build_number: int, build_url: str, latest: bool, sound: bool) -> None:
    """Start monitor UI

    Args:
        profile: The profile/account to use
        job: The job this build is under
        build_number: The build number to get info on
        build_url: The build url to get info on
        latest: Option to get the latest build
        sound: Option to play a sound when the build status changes

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        yj_obj.build.monitor(build_url=build_url, job_url=job, build_number=build_number, latest=latest, sound=sound)
    else:
        yj_obj.build.monitor(build_url=build_url, job_name=job, build_number=build_number, latest=latest, sound=sound)


@log_to_history
def parameters(opt_pretty: bool, opt_yaml: bool, opt_xml: bool, opt_toml: bool, profile: str, opt_list: bool, job: str,
               build_number: int, build_url: str, latest: bool) -> None:
    """Get build parameters

    Args:
        profile: The profile/account to use
        opt_list: Option to list all stages without details
        job: The job name under which the build is located
        build_number: The build number
        build_url: The build URL
        latest: Option to get the latest build

    Returns:
        None
    """
    if job and not build_number and not latest:
        click.echo(
            click.style('INPUT ERROR: For job, either specify --number or --latest. See --help',
                        fg='bright_red',
                        bold=True))
        sys.exit(1)

    yj_obj = cu.config_yo_jenkins(profile)

    if _verify_build_url_get_job_format(build_url=build_url, job=job):
        data, data_list = yj_obj.build.parameters(build_url=build_url,
                                                  job_url=job,
                                                  build_number=build_number,
                                                  latest=latest)
    else:
        data, data_list = yj_obj.build.parameters(build_url=build_url,
                                                  job_name=job,
                                                  build_number=build_number,
                                                  latest=latest)
    data = data_list if opt_list else data
    cu.standard_out(data, opt_pretty, opt_yaml, opt_xml, opt_toml)



"""Step class definition"""

import logging
from typing import Dict

from yojenkins.utility.utility import fail_out

# Getting the logger reference
logger = logging.getLogger()


class Step():
    """TODO Step"""

    def __init__(self, rest: object) -> None:
        """Object constructor method, called at object creation

        Args:
            None

        Returns:
            None
        """
        self.rest = rest

    def info(self, step_url=str) -> Dict:
        """Get the information of the specified step

        Details: Step URL can be gotten with stage_step_list.
                 Format: <SERVER URL>/<step_url>

        Args:
            step_url : The direct URL of the step

        Returns:
            Step information
        """
        request_url = f'{step_url.strip("/")}'
        return_content = self.rest.request(request_url, 'get', is_endpoint=True)[0]
        if not return_content:
            fail_out(f'Failed to get step info for: {request_url}')

        return return_content
